@@ -8,28 +8,26 @@ FTPFS is a filesystem for accessing an FTP server (uses ftplib in standard libra
 
 __all__ = ['FTPFS']
 
+import sys 
+
 import fs
 from fs.base import *
-from fs.path import pathsplit
+from fs.errors import *
+from fs.path import pathsplit, abspath, dirname, recursepath, normpath, pathjoin, isbase
 
 from ftplib import FTP, error_perm, error_temp, error_proto, error_reply
 
 try:
-    from ftplib import _GLOBAL_DEFAULT_TIMEOUT
-    _FTPLIB_TIMEOUT = True
+    from ftplib import _GLOBAL_DEFAULT_TIMEOUT    
 except ImportError:
-    _GLOBAL_DEFAULT_TIMEOUT = None
-    _FTPLIB_TIMEOUT = False
+    _GLOBAL_DEFAULT_TIMEOUT = object()    
 
 import threading
-from time import sleep
 import datetime
-import re
+import calendar
+
 from socket import error as socket_error
-try:
-    from functools import wraps
-except ImportError:
-    wraps = lambda f: lambda f: f
+from fs.local_functools import wraps
 
 try:
     from cStringIO import StringIO
@@ -37,7 +35,6 @@ except ImportError:
     from StringIO import StringIO
 
 import time
-import sys
 
 
 # -----------------------------------------------
@@ -310,7 +307,7 @@ class FTPListDataParser(object):
                         result.mtime = self._guess_time(month, mday, hour, minute)
                     elif j - i >= 4:
                         year = long(buf[i:j])
-                        result.mtimetype = MTIME_TYPE.REMOTE_DAY
+                        result.mtime_type = MTIME_TYPE.REMOTE_DAY
                         result.mtime = self._get_mtime(year, month, mday)
                     else:
                         break
@@ -403,7 +400,7 @@ class FTPListDataParser(object):
 
             minute = long(buf[i:j])
 
-            result.mtimetype = MTIME_TYPE.REMOTE_MINUTE
+            result.mtime_type = MTIME_TYPE.REMOTE_MINUTE
             result.mtime = self._get_mtime(year, month, mday, hour, minute)
 
         except IndexError:
@@ -483,19 +480,84 @@ class FTPListDataParser(object):
             j = _skip(buf, j, ' ')
 
             result.name = buf[j:]
-            result.mtimetype = MTIME_TYPE.REMOTE_MINUTE
+            result.mtime_type = MTIME_TYPE.REMOTE_MINUTE
             result.mtime = self._get_mtime(year, month, mday, hour, minute)
         except IndexError:
             pass
 
         return result
 
+class FTPMlstDataParser(object):
+    """
+    An ``FTPMlstDataParser`` object can be used to parse one or more lines
+    that were retrieved by an FTP ``MLST`` or ``MLSD`` command that was sent
+    to a remote server.
+    """
+    def __init__(self):
+        pass
+
+    def parse_line(self, ftp_list_line):
+        """
+        Parse a line from an FTP ``MLST`` or ``MLSD`` command.
+
+        :Parameters:
+            ftp_list_line : str
+                The line of output
+
+        :rtype: `FTPListData`
+        :return: An `FTPListData` object describing the parsed line, or
+                 ``None`` if the line could not be parsed. Note that it's
+                 possible for this method to return a partially-filled
+                 `FTPListData` object (e.g., one without a mtime).
+        """
+        result = FTPListData(ftp_list_line)
+        # pull out the name
+        parts = ftp_list_line.partition(' ')
+        result.name = parts[2]
+
+        # parse the facts
+        if parts[0][-1] == ';':
+            for fact in parts[0][:-1].split(';'):
+                parts = fact.partition('=')
+                factname = parts[0].lower()
+                factvalue = parts[2]
+                if factname == 'unique':
+                    if factvalue == "0g0" or factvalue == "0g1":
+                        # Matrix FTP server sometimes returns bogus "unique" facts
+                        result.id_type = ID_TYPE.UNKNOWN
+                    else:
+                        result.id_type = ID_TYPE.FULL
+                    result.id = factvalue
+                elif factname == 'modify':
+                    result.mtime_type = MTIME_TYPE.LOCAL
+                    result.mtime = calendar.timegm((int(factvalue[0:4]),
+                                                     int(factvalue[4:6]),
+                                                     int(factvalue[6:8]),
+                                                     int(factvalue[8:10]),
+                                                     int(factvalue[10:12]),
+                                                     int(factvalue[12:14]),
+                                                     0, 0, 0))
+                elif factname == 'size':
+                    result.size = long(factvalue)
+                elif factname == 'sizd':
+                    # some FTP servers report directory size with sizd
+                    result.size = long(factvalue)
+                elif factname == 'type':
+                    if factvalue.lower() == 'file':
+                        result.try_retr = True
+                    elif factvalue.lower() in ['dir', 'cdir', 'pdir']:
+                        result.try_cwd = True
+                    else:
+                        # dunno if it's file or directory
+                        result.try_retr = True
+                        result.try_cwd = True
+        return result
 
 # ---------------------------------------------------------------------------
 # Public Functions
 # ---------------------------------------------------------------------------
 
-def parse_ftp_list_line(ftp_list_line):
+def parse_ftp_list_line(ftp_list_line, is_mlst=False):
     """
     Convenience function that instantiates an `FTPListDataParser` object
     and passes ``ftp_list_line`` to the object's ``parse_line()`` method,
@@ -511,7 +573,10 @@ def parse_ftp_list_line(ftp_list_line):
              possible for this method to return a partially-filled
              `FTPListData` object (e.g., one without a name).
     """
-    return FTPListDataParser().parse_line(ftp_list_line)
+    if is_mlst:
+        return FTPMlstDataParser().parse_line(ftp_list_line)
+    else:
+        return FTPListDataParser().parse_line(ftp_list_line)
 
 # ---------------------------------------------------------------------------
 # Private Functions
@@ -525,7 +590,19 @@ def _skip(s, i, c):
     return i
 
 
-
+def fileftperrors(f):    
+    @wraps(f)
+    def deco(self, *args, **kwargs):
+        self._lock.acquire()
+        try:                        
+            try:
+                ret = f(self, *args, **kwargs)
+            except Exception, e:                
+                self.ftpfs._translate_exception(args[0] if args else '', e)                        
+        finally:
+            self._lock.release()
+        return ret
+    return deco
 
 
 
@@ -533,44 +610,50 @@ class _FTPFile(object):
 
     """ A file-like that provides access to a file being streamed over ftp."""
 
+    blocksize = 1024 * 64
+
     def __init__(self, ftpfs, ftp, path, mode):
         if not hasattr(self, '_lock'):
-            self._lock = threading.RLock()
+            self._lock = threading.RLock()        
         self.ftpfs = ftpfs
         self.ftp = ftp
-        self.path = path
+        self.path = normpath(path)
         self.mode = mode
         self.read_pos = 0
         self.write_pos = 0
         self.closed = False
+        self.file_size = None
         if 'r' in mode or 'a' in mode:
             self.file_size = ftpfs.getsize(path)
         self.conn = None
 
-        path = _encode(path)
-        #self._lock = ftpfs._lock
+        self._start_file(mode, _encode(self.path))
 
+    @fileftperrors
+    def _start_file(self, mode, path):
+        self.read_pos = 0
+        self.write_pos = 0
         if 'r' in mode:
             self.ftp.voidcmd('TYPE I')
-            self.conn = ftp.transfercmd('RETR '+path, None)
+            self.conn = self.ftp.transfercmd('RETR ' + path, None)
 
-        elif 'w' in mode or 'a' in mode:
+        else:#if 'w' in mode or 'a' in mode:
             self.ftp.voidcmd('TYPE I')
             if 'a' in mode:
                 self.write_pos = self.file_size
-                self.conn = self.ftp.transfercmd('APPE '+path)
+                self.conn = self.ftp.transfercmd('APPE ' + path)
             else:
-                self.conn = self.ftp.transfercmd('STOR '+path)
+                self.conn = self.ftp.transfercmd('STOR ' + path)
 
-    @synchronize
+    @fileftperrors
     def read(self, size=None):
         if self.conn is None:
             return ''
 
         chunks = []
-        if size is None:
+        if size is None or size < 0:
             while 1:
-                data = self.conn.recv(4096)
+                data = self.conn.recv(self.blocksize)
                 if not data:
                     self.conn.close()
                     self.conn = None
@@ -582,7 +665,7 @@ class _FTPFile(object):
 
         remaining_bytes = size
         while remaining_bytes:
-            read_size = min(remaining_bytes, 4096)
+            read_size = min(remaining_bytes, self.blocksize)
             data = self.conn.recv(read_size)
             if not data:
                 self.conn.close()
@@ -595,34 +678,36 @@ class _FTPFile(object):
 
         return ''.join(chunks)
 
-    @synchronize
+    @fileftperrors
     def write(self, data):
 
         data_pos = 0
         remaining_data = len(data)
 
         while remaining_data:
-            chunk_size = min(remaining_data, 4096)
+            chunk_size = min(remaining_data, self.blocksize)
             self.conn.sendall(data[data_pos:data_pos+chunk_size])
             data_pos += chunk_size
             remaining_data -= chunk_size
             self.write_pos += chunk_size
 
+
     def __enter__(self):
         return self
 
     def __exit__(self,exc_type,exc_value,traceback):
         self.close()
 
-    @synchronize
+    @fileftperrors
     def flush(self):
-        return
+        self.ftpfs._on_file_written(self.path)
 
+    @fileftperrors
     def seek(self, pos, where=fs.SEEK_SET):
         # Ftp doesn't support a real seek, so we close the transfer and resume
         # it at the new position with the REST command
         # I'm not sure how reliable this method is!
-        if not self.file_size:
+        if self.file_size is None:
             raise ValueError("Seek only works with files open for read")
 
         self._lock.acquire()
@@ -658,28 +743,63 @@ class _FTPFile(object):
 
         #raise UnsupportedError('ftp seek')
 
-    @synchronize
+    @fileftperrors
     def tell(self):
         if 'r' in self.mode:
             return self.read_pos
         else:
             return self.write_pos
 
-    @synchronize
-    def close(self):
-        if self.conn is not None:
+    @fileftperrors
+    def truncate(self, size=None):
+        self.ftpfs._on_file_written(self.path)
+        # Inefficient, but I don't know how else to implement this
+        if size is None:        
+            size = self.tell()
+         
+        if self.conn is not None:           
             self.conn.close()
-            self.conn = None
-            self.ftp.voidresp()
-        if self.ftp is not None:
-            self.ftp.close()
-        self.closed = True
-        if 'w' in self.mode or 'a' in self.mode:
+        self.close()
+                
+        read_f = None        
+        try:
+            read_f = self.ftpfs.open(self.path, 'rb')
+            data = read_f.read(size)
+        finally:
+            if read_f is not None:
+                read_f.close()           
+            
+        self.ftp = self.ftpfs._open_ftp()        
+        self.mode = 'w'
+        self.__init__(self.ftpfs, self.ftp, _encode(self.path), self.mode)
+        #self._start_file(self.mode, self.path)        
+        self.write(data)
+        if len(data) < size:
+            self.write('\0' * (size - len(data)))
+        
+
+    @fileftperrors
+    def close(self):
+        if 'w' in self.mode or 'a' in self.mode or '+' in self.mode:
             self.ftpfs._on_file_written(self.path)
+        if self.conn is not None:
+            try:
+                self.conn.close()
+                self.conn = None
+                self.ftp.voidresp()
+            except error_temp, error_perm:
+                pass
+        if self.ftp is not None:
+            try:
+                self.ftp.close()
+            except error_temp, error_perm:
+                pass
+        self.closed = True        
 
     def __iter__(self):
         return self.next()
 
+    @synchronize
     def next(self):
         """ Line iterator
 
@@ -711,9 +831,7 @@ class _FTPFile(object):
                     yield line
                     append(c)
 
-
-
-def ftperrors(f):
+def ftperrors(f):    
     @wraps(f)
     def deco(self, *args, **kwargs):
         self._lock.acquire()
@@ -725,30 +843,49 @@ def ftperrors(f):
                 except Exception, e:
                     self._translate_exception(args[0] if args else '', e)
             finally:
-                self._leave_dircache()
+                self._leave_dircache()                
         finally:
             self._lock.release()
-        if not self.use_dircache:
-            self.clear_dircache()
         return ret
     return deco
 
 
-def _encode(s):
+def _encode(s):    
     if isinstance(s, unicode):
         return s.encode('utf-8')
     return s
 
+class _DirCache(dict):
+    def __init__(self): 
+        super(_DirCache, self).__init__()       
+        self.count = 0
+    
+    def addref(self):
+        self.count += 1
+        return self.count
+        
+    def decref(self):
+        self.count -= 1
+        return self.count
 
-class FTPFS(FS):
+class FTPFS(FS):    
+    
+    _meta = { 'thread_safe' : True,
+              'network' : True,
+              'virtual': False,
+              'read_only' : False,
+              'unicode_paths' : True,
+              'case_insensitive_paths' : False,
+              'atomic.move' : True,
+              'atomic.copy' : True,              
+              'atomic.makedir' : True,
+              'atomic.rename' : True,
+              'atomic.setcontents' : False,
+              'file.read_and_write' : False,
+              }
 
-    _locals = threading.local()
-
-    def __init__(self, host='', user='', passwd='', acct='', timeout=_GLOBAL_DEFAULT_TIMEOUT,
-                 port=21,
-                 dircache=True,
-                 max_buffer_size=128*1024*1024):
-        """ Connect to a FTP server.
+    def __init__(self, host='', user='', passwd='', acct='', timeout=_GLOBAL_DEFAULT_TIMEOUT, port=21, dircache=True):
+        """Connect to a FTP server.
         
         :param host: Host to connect to
         :param user: Username, or a blank string for anonymous
@@ -757,11 +894,9 @@ class FTPFS(FS):
         :param timeout: Timeout in seconds
         :param port: Port to connection (default is 21)
         :param dircache: If True then directory information will be cached,
-        which will speed up operations such as getinfo, isdi, isfile, but
-        changes to the ftp file structure will not be visible until
-        `~fs.ftpfs.FTPFS.clear_dircache` is called        
-        :param dircache: If True directory information will be cached for fast access
-        :param max_buffer_size: Number of bytes to hold before blocking write operations
+            speeding up operations such as `getinfo`, `isdir`, `isfile`, but
+            changes to the ftp file structure will not be visible until
+            :meth:`~fs.ftpfs.FTPFS.clear_dircache` is called                
 
         """
 
@@ -773,77 +908,108 @@ class FTPFS(FS):
         self.passwd = passwd
         self.acct = acct
         self.timeout = timeout
+        self.default_timeout = timeout is _GLOBAL_DEFAULT_TIMEOUT
+        self.use_dircache = dircache            
 
-        self.use_dircache = dircache
-        self.get_dircache()
-
-        self.max_buffer_size = max_buffer_size
-
+        self.use_mlst = False
+        self._lock = threading.RLock()
+        self._init_dircache()
+                    
         self._cache_hint = False
-        self._locals._ftp = None
-        self._thread_ftps = set()
-        self.ftp
+        try:        
+            self.ftp
+        except FSError:
+            self.closed = True
+            raise
+
+    def _init_dircache(self):        
+        self.dircache = _DirCache()
 
     @synchronize
-    def cache_hint(self, enabled):
-        self._cache_hint = enabled
+    def cache_hint(self, enabled):        
+        self._cache_hint = bool(enabled)
 
-    @synchronize
     def _enter_dircache(self):
-        self.get_dircache()
-        count = getattr(self._locals, '_dircache_count', 0)
-        count += 1
-        self._locals._dircache_count = count
+        self.dircache.addref()        
 
-    @synchronize
     def _leave_dircache(self):
-        self._locals._dircache_count -= 1
-        if not self._locals._dircache_count and not self._cache_hint:
+        self.dircache.decref()
+        if self.use_dircache:
+            if not self.dircache.count and not self._cache_hint:
+                self.clear_dircache()
+        else:
             self.clear_dircache()
-        assert self._locals._dircache_count >= 0, "dircache count should never be negative"
-
-    @synchronize
-    def get_dircache(self):
-        dircache = getattr(self._locals, '_dircache', None)
-        if dircache is None:
-            dircache = {}
-            self._locals._dircache = dircache
-            self._locals._dircache_count = 0
-        return dircache
-
+        assert self.dircache.count >= 0, "dircache count should never be negative"                    
+    
     @synchronize
     def _on_file_written(self, path):
-        self.clear_dircache(dirname(path))
-
+        self.refresh_dircache(dirname(path))
 
     @synchronize
     def _readdir(self, path):
-
-        dircache = self.get_dircache()
-        dircache_count = self._locals._dircache_count
-
-        if dircache_count:
-            cached_dirlist = dircache.get(path)
+        path = abspath(normpath(path))
+        if self.dircache.count:
+            cached_dirlist = self.dircache.get(path)
             if cached_dirlist is not None:
                 return cached_dirlist
         dirlist = {}
 
-        parser = FTPListDataParser()
+        def _get_FEAT(ftp):
+            features = dict()
+            try:
+                response = ftp.sendcmd("FEAT")
+                if response[:3] == "211":
+                    for line in response.splitlines()[1:]:
+                        if line[3] == "211":
+                            break
+                        if line[0] != ' ':
+                            break
+                        parts = line[1:].partition(' ')
+                        features[parts[0].upper()] = parts[2]
+            except error_perm:
+                # some FTP servers may not support FEAT
+                pass
+            return features
 
         def on_line(line):
-            #print repr(line)
             if not isinstance(line, unicode):
-                line = line.decode('utf-8')
-            info = parser.parse_line(line)
+                line = line.decode('utf-8')                
+            info = parse_ftp_list_line(line, self.use_mlst)
             if info:
                 info = info.__dict__
-                dirlist[info['name']] = info
+                if info['name'] not in ('.', '..'):
+                    dirlist[info['name']] = info
 
         try:
-            self.ftp.dir(_encode(path), on_line)
+            encoded_path = _encode(path)
+            ftp_features = _get_FEAT(self.ftp)
+            if 'MLST' in ftp_features:
+                self.use_mlst = True
+                try:
+                    # only request the facts we need
+                    self.ftp.sendcmd("OPTS MLST type;unique;size;modify;")
+                except error_perm:
+                    # some FTP servers don't support OPTS MLST
+                    pass
+                # need to send MLST first to discover if it's file or dir
+                response = self.ftp.sendcmd("MLST " + encoded_path)
+                lines = response.splitlines()
+                if lines[0][:3] == "250":
+                    list_line = lines[1]
+                    # MLST line is preceded by space
+                    if list_line[0] == ' ':
+                        on_line(list_line[1:])
+                    else: # Matrix FTP server has bug
+                        on_line(list_line)
+                # if it's a dir, then we can send a MLSD
+                if dirlist[dirlist.keys()[0]]['try_cwd']:
+                    dirlist = {}
+                    self.ftp.retrlines("MLSD " + encoded_path, on_line)
+            else:
+                self.ftp.dir(encoded_path, on_line)
         except error_reply:
             pass
-        dircache[path] = dirlist
+        self.dircache[path] = dirlist
 
         return dirlist
 
@@ -855,16 +1021,28 @@ class FTPFS(FS):
         :param path: Path of directory to clear cache for, or all directories if
         None (the default)
 
-        """
-        dircache = self.get_dircache()
+        """        
+
         if not paths:
-            dircache.clear()
-        else:
-            for path in paths:
-                dircache.pop(path, None)
+            self.dircache.clear()
+        else:            
+            dircache = self.dircache
+            paths = [normpath(abspath(path)) for path in paths]
+            for cached_path in dircache.keys():                
+                for path in paths:
+                    if isbase(cached_path, path):
+                        dircache.pop(cached_path, None)
+                        break
+    
+    @synchronize
+    def refresh_dircache(self, *paths):                                     
+        for path in paths:
+            path = abspath(normpath(path))
+            self.dircache.pop(path, None)      
 
     @synchronize
     def _check_path(self, path):
+        path = normpath(path)
         base, fname = pathsplit(abspath(path))
         dirlist = self._readdir(base)
         if fname and fname not in dirlist:
@@ -872,46 +1050,47 @@ class FTPFS(FS):
         return dirlist, fname
 
     def _get_dirlist(self, path):
+        path = normpath(path)
         base, fname = pathsplit(abspath(path))
         dirlist = self._readdir(base)
         return dirlist, fname
 
 
-    @synchronize
+    @ftperrors
     def get_ftp(self):
-        if getattr(self._locals, '_ftp', None) is None:
-            self._locals._ftp = self._open_ftp()
-        ftp = self._locals._ftp
-        self._thread_ftps.add(ftp)
-        return self._locals._ftp
-    @synchronize
-    def set_ftp(self, ftp):
-        self._locals._ftp = ftp
-    ftp = property(get_ftp, set_ftp)
+        if self.closed:
+            return None        
+        if not getattr(self, '_ftp', None):
+            self._ftp = self._open_ftp()
+        return self._ftp          
+      
+    ftp = property(get_ftp)
 
-    @synchronize
+    @ftperrors
     def _open_ftp(self):
         try:
-            ftp = FTP()
-            if _FTPLIB_TIMEOUT:
-                ftp.connect(self.host, self.port, self.timeout)
-            else:
+            ftp = FTP()            
+            if self.default_timeout or sys.version_info < (2,6,):
                 ftp.connect(self.host, self.port)
+            else:                
+                ftp.connect(self.host, self.port, self.timeout)
             ftp.login(self.user, self.passwd, self.acct)
         except socket_error, e:
             raise RemoteConnectionError(str(e), details=e)
         return ftp
 
     def __getstate__(self):
-        state = super(FTPFS, self).__getstate__()
-        del state["_thread_ftps"]
+        state = super(FTPFS, self).__getstate__()        
+        del state['_lock']
+        state.pop('_ftp', None)
         return state
 
     def __setstate__(self,state):
-        super(FTPFS, self).__setstate__(state)
-        self._thread_ftps = set()
-        self.ftp
-
+        super(FTPFS, self).__setstate__(state)        
+        self._init_dircache()
+        self._lock = threading.RLock()
+        #self._ftp = None
+        #self.ftp
 
     def __str__(self):
         return '<FTPFS %s>' % self.host
@@ -930,52 +1109,84 @@ class FTPFS(FS):
         """
 
         if isinstance(exception, socket_error):
+            self._ftp = None
             raise RemoteConnectionError(str(exception), details=exception)
 
         elif isinstance(exception, error_temp):
             code, message = str(exception).split(' ', 1)
-            raise RemoteConnectionError(str(exception), path=path, msg="FTP error: %s (see details)" % str(exception), details=exception)
+            self._ftp = None
+            raise RemoteConnectionError(str(exception), path=path, msg="FTP error: %s" % str(exception), details=exception)
 
         elif isinstance(exception, error_perm):
             code, message = str(exception).split(' ', 1)
             code = int(code)
             if code == 550:
-                raise ResourceNotFoundError(path)
+                pass                
             if code == 552:
                 raise StorageSpaceError
-            raise PermissionDeniedError(str(exception), path=path, msg="FTP error: %s (see details)" % str(exception), details=exception)
+            raise PermissionDeniedError(str(exception), path=path, msg="FTP error: %s" % str(exception), details=exception)
 
         raise exception
 
     @ftperrors
     def close(self):
-        for ftp in self._thread_ftps:
-            ftp.close()
+        if not self.closed:
+            try:
+                self.ftp.close()
+            except FSError:
+                pass        
+            self.closed = True
 
-        self._thread_ftps.clear()
-        self.closed = True
+    def getpathurl(self, path, allow_none=False):
+        path = normpath(path)
+        credentials = '%s:%s' % (self.user, self.passwd)        
+        if credentials == ':':            
+            url = 'ftp://%s%s' % (self.host.rstrip('/'), abspath(path))
+        else:
+            url = 'ftp://%s@%s%s' % (credentials, self.host.rstrip('/'), abspath(path))        
+        return url             
 
     @ftperrors
     def open(self, path, mode='r'):
-        mode = mode.lower()
-        if 'r' in mode:
+        path = normpath(path)
+        mode = mode.lower()        
+        if self.isdir(path):
+            raise ResourceInvalidError(path)        
+        if 'r' in mode or 'a' in mode:
             if not self.isfile(path):
                 raise ResourceNotFoundError(path)
-        if 'w' in mode or 'a' in mode:
-            self.clear_dircache(dirname(path))
+        if 'w' in mode or 'a' in mode or '+' in mode:
+            self.refresh_dircache(dirname(path))
         ftp = self._open_ftp()
-        f = _FTPFile(self, ftp, path, mode)
-        return f
+        f = _FTPFile(self, ftp, normpath(path), mode) 
+        return f 
+    
+    @ftperrors
+    def setcontents(self, path, data, chunk_size=1024*64):        
+        path = normpath(path)        
+        if isinstance(data, basestring):
+            data = StringIO(data)        
+        self.refresh_dircache(dirname(path))        
+        self.ftp.storbinary('STOR %s' % _encode(path), data, blocksize=chunk_size)        
+        
+    @ftperrors
+    def getcontents(self, path):
+        path = normpath(path) 
+        contents = StringIO()                            
+        self.ftp.retrbinary('RETR %s' % _encode(path), contents.write, blocksize=1024*64)            
+        return contents.getvalue()        
 
     @ftperrors
     def exists(self, path):
+        path = normpath(path)
         if path in ('', '/'):
-            return True
+            return True        
         dirlist, fname = self._get_dirlist(path)
         return fname in dirlist
 
     @ftperrors
     def isdir(self, path):
+        path = normpath(path)
         if path in ('', '/'):
             return True
         dirlist, fname = self._get_dirlist(path)
@@ -986,6 +1197,7 @@ class FTPFS(FS):
 
     @ftperrors
     def isfile(self, path):
+        path = normpath(path)
         if path in ('', '/'):
             return False
         dirlist, fname = self._get_dirlist(path)
@@ -995,8 +1207,9 @@ class FTPFS(FS):
         return not info['try_cwd']
 
     @ftperrors
-    def listdir(self, path="./", wildcard=None, full=False, absolute=False, dirs_only=False, files_only=False):
+    def listdir(self, path="./", wildcard=None, full=False, absolute=False, dirs_only=False, files_only=False):                
         path = normpath(path)
+        #self.clear_dircache(path)
         if not self.exists(path):
             raise ResourceNotFoundError(path)
         if not self.isdir(path):
@@ -1005,24 +1218,50 @@ class FTPFS(FS):
 
         return self._listdir_helper(path, paths, wildcard, full, absolute, dirs_only, files_only)
 
+    @ftperrors
+    def listdirinfo(self, path="./",
+                          wildcard=None,
+                          full=False,
+                          absolute=False,
+                          dirs_only=False,
+                          files_only=False):
+        path = normpath(path)
+        def getinfo(p):
+            try:
+                if full or absolute:
+                    return self.getinfo(p)
+                else:
+                    return self.getinfo(pathjoin(path,p))
+            except FSError:
+                return {}
+
+        return [(p, getinfo(p))
+                    for p in self.listdir(path,                                          
+                                          wildcard=wildcard,
+                                          full=full,
+                                          absolute=absolute,
+                                          dirs_only=dirs_only,
+                                          files_only=files_only)]
 
     @ftperrors
     def makedir(self, path, recursive=False, allow_recreate=False):
+        path = normpath(path)
         if path in ('', '/'):
             return
         def checkdir(path):
-            self.clear_dircache(dirname(path), path)
-            try:
-                self.ftp.mkd(_encode(path))
-            except error_reply:
-                return
-            except error_perm, e:
-                if recursive or allow_recreate:
+            if not self.isdir(path):
+                self.clear_dircache(dirname(path))
+                try:
+                    self.ftp.mkd(_encode(path))
+                except error_reply:
                     return
-                if str(e).split(' ', 1)[0]=='550':
-                    raise DestinationExistsError(path)
-                else:
-                    raise
+                except error_perm, e:
+                    if recursive or allow_recreate:
+                        return
+                    if str(e).split(' ', 1)[0]=='550':
+                        raise DestinationExistsError(path)
+                    else:
+                        raise
         if recursive:
             for p in recursepath(path):
                 checkdir(p)
@@ -1036,27 +1275,27 @@ class FTPFS(FS):
                     if self.isfile(path):
                         raise ResourceInvalidError(path)
                     raise DestinationExistsError(path)
-            checkdir(path)
+            checkdir(path)        
 
     @ftperrors
-    def remove(self, path):
+    def remove(self, path):        
         if not self.exists(path):
             raise ResourceNotFoundError(path)
         if not self.isfile(path):
             raise ResourceInvalidError(path)
-        self.clear_dircache(dirname(path))
-        self.ftp.delete(_encode(path))
+        self.refresh_dircache(dirname(path))        
+        self.ftp.delete(_encode(path))        
 
     @ftperrors
     def removedir(self, path, recursive=False, force=False):
-
+        path = abspath(normpath(path))
         if not self.exists(path):
             raise ResourceNotFoundError(path)
         if self.isfile(path):
             raise ResourceInvalidError(path)
 
         if not force:
-            for checkpath in self.listdir(path):
+            for _checkpath in self.listdir(path):
                 raise DirectoryNotEmptyError(path)
         try:
             if force:
@@ -1068,7 +1307,7 @@ class FTPFS(FS):
                             self.removedir(rpath, force=force)
                     except FSError:
                         pass
-            self.clear_dircache(dirname(path), path)
+            self.clear_dircache(dirname(path))
             self.ftp.rmd(_encode(path))
         except error_reply:
             pass
@@ -1077,19 +1316,21 @@ class FTPFS(FS):
                 self.removedir(dirname(path), recursive=True)
             except DirectoryNotEmptyError:
                 pass
+        self.clear_dircache(dirname(path), path)
 
     @ftperrors
-    def rename(self, src, dst):
-        self.clear_dircache(dirname(src), dirname(dst), src, dst)
+    def rename(self, src, dst):        
         try:
-            self.ftp.rename(_encode(src), _encode(dst))
+            self.refresh_dircache(dirname(src), dirname(dst))
+            self.ftp.rename(_encode(src), _encode(dst))            
         except error_perm, exception:
             code, message = str(exception).split(' ', 1)
             if code == "550":
                 if not self.exists(dirname(dst)):
                     raise ParentDirectoryMissingError(dst)
+            raise
         except error_reply:
-            pass
+            pass        
 
     @ftperrors
     def getinfo(self, path):
@@ -1105,7 +1346,7 @@ class FTPFS(FS):
     def getsize(self, path):
 
         size = None
-        if self._locals._dircache_count:
+        if self.dircache.count:
             dirlist, fname = self._check_path(path)
             size = dirlist[fname].get('size')
 
@@ -1123,6 +1364,10 @@ class FTPFS(FS):
 
     @ftperrors
     def desc(self, path):
+        path = normpath(path)
+        url = self.getpathurl(path, allow_none=True)
+        if url:
+            return url
         dirlist, fname = self._check_path(path)
         if fname not in dirlist:
             raise ResourceNotFoundError(path)
@@ -1130,27 +1375,48 @@ class FTPFS(FS):
 
     @ftperrors
     def move(self, src, dst, overwrite=False, chunk_size=16384):
-
         if not overwrite and self.exists(dst):
             raise DestinationExistsError(dst)
-        self.clear_dircache(dirname(src), dirname(dst))
+        #self.refresh_dircache(dirname(src), dirname(dst))        
         try:
             self.rename(src, dst)
-        except error_reply:
-            pass
         except:
-            self.copy(src, dst)
+            self.copy(src, dst, overwrite=overwrite)
             self.remove(src)
+        finally:
+            self.refresh_dircache(src, dirname(src), dst, dirname(dst))                    
+    
+    @ftperrors        
+    def copy(self, src, dst, overwrite=False, chunk_size=1024*64):        
+        if not self.isfile(src):
+            if self.isdir(src):
+                raise ResourceInvalidError(src, msg="Source is not a file: %(path)s")
+            raise ResourceNotFoundError(src)
+        if not overwrite and self.exists(dst):
+            raise DestinationExistsError(dst)
+    
+        dst = normpath(dst)
+        src_file = None            
+        try:
+            src_file = self.open(src, "rb")
+            ftp = self._open_ftp()  
+            ftp.voidcmd('TYPE I')              
+            ftp.storbinary('STOR %s' % _encode(normpath(dst)), src_file, blocksize=chunk_size)                                                    
+        finally:
+            self.refresh_dircache(dirname(dst))            
+            if src_file is not None:
+                src_file.close()   
+                         
 
     @ftperrors
     def movedir(self, src, dst, overwrite=False, ignore_errors=False, chunk_size=16384):
-        self.clear_dircache(src, dst, dirname(src), dirname(dst))
-        super(FTPFS, self).movedir(src, dst, overwrite, ignore_errors, chunk_size)
+        self.clear_dircache(dirname(src), dirname(dst))        
+        super(FTPFS, self).movedir(src, dst, overwrite, ignore_errors, chunk_size)        
 
     @ftperrors
     def copydir(self, src, dst, overwrite=False, ignore_errors=False, chunk_size=16384):
-        self.clear_dircache(src, dst, dirname(src), dirname(dst))
-        super(FTPFS, self).copydir(src, dst, overwrite, ignore_errors, chunk_size)
+        self.clear_dircache(dirname(dst))        
+        super(FTPFS, self).copydir(src, dst, overwrite, ignore_errors, chunk_size)        
 
 
 if __name__ == "__main__":
