@@ -9,9 +9,10 @@ interface for objects stored in Amazon Simple Storage Service (S3).
 
 """
 
-import time
+import os
 import datetime
 import tempfile
+from fnmatch import fnmatch
 import stat as statinfo
 
 import boto.s3.connection
@@ -19,7 +20,10 @@ from boto.s3.prefix import Prefix
 from boto.exception import S3ResponseError
 
 from fs.base import *
+from fs.path import *
+from fs.errors import *
 from fs.remote import *
+from fs.filelike import LimitBytesFile
 
 
 # Boto is not thread-safe, so we need to use a per-thread S3 connection.
@@ -30,7 +34,6 @@ else:
         def __init__(self):
             self._map = {}
         def __getattr__(self,attr):
-
             try:
                 return self._map[(threading.currentThread(),attr)]
             except KeyError:
@@ -52,16 +55,29 @@ class S3FS(FS):
     or flushed.
     """
 
+    _meta = { 'thread_safe' : True,
+              'virtual': False,
+              'read_only' : False,
+              'unicode_paths' : True,
+              'case_insensitive_paths' : False,
+              'network' : True,
+              'atomic.move' : True,
+              'atomic.copy' : True,
+              'atomic.makedir' : True,
+              'atomic.rename' : False,
+              'atomic.setconetns' : True
+              }
+
     class meta:
         PATH_MAX = None
         NAME_MAX = None
 
-    def __init__(self, bucket, prefix="", aws_access_key=None, aws_secret_key=None, separator="/", thread_synchronize=True,key_sync_timeout=1):
+    def __init__(self, bucket, prefix="", aws_access_key=None, aws_secret_key=None, separator="/", thread_synchronize=True, key_sync_timeout=1):
         """Constructor for S3FS objects.
 
         S3FS objects require the name of the S3 bucket in which to store
         files, and can optionally be given a prefix under which the files
-        shoud be stored.  The AWS public and private keys may be specified
+        should be stored.  The AWS public and private keys may be specified
         as additional arguments; if they are not specified they will be
         read from the two environment variables AWS_ACCESS_KEY_ID and
         AWS_SECRET_ACCESS_KEY.
@@ -86,6 +102,14 @@ class S3FS(FS):
             prefix = prefix[1:]
         if not prefix.endswith(separator) and prefix != "":
             prefix = prefix + separator
+        if isinstance(prefix,unicode):
+            prefix = prefix.encode("utf8")
+        if aws_access_key is None:
+            if "AWS_ACCESS_KEY_ID" not in os.environ:
+                raise CreateFailedError("AWS_ACCESS_KEY_ID not set")
+        if aws_secret_key is None:
+            if "AWS_SECRET_ACCESS_KEY" not in os.environ:
+                raise CreateFailedError("AWS_SECRET_ACCESS_KEY not set")
         self._prefix = prefix
         self._tlocal = thread_local()
         super(S3FS, self).__init__(thread_synchronize=thread_synchronize)
@@ -95,24 +119,30 @@ class S3FS(FS):
 
     def _s3conn(self):
         try:
-            return self._tlocal.s3conn
+            (c,ctime) = self._tlocal.s3conn
+            if time.time() - ctime > 60:
+                raise AttributeError
+            return c
         except AttributeError:
             c = boto.s3.connection.S3Connection(*self._access_keys)
-            self._tlocal.s3conn = c
+            self._tlocal.s3conn = (c,time.time())
             return c
     _s3conn = property(_s3conn)
 
     def _s3bukt(self):
         try:
-            return self._tlocal.s3bukt
+            (b,ctime) = self._tlocal.s3bukt
+            if time.time() - ctime > 60:
+                raise AttributeError
+            return b
         except AttributeError:
             try:
-                b = self._s3conn.get_bucket(self._bucket_name)
+                b = self._s3conn.get_bucket(self._bucket_name, validate=True)
             except S3ResponseError, e:
                 if "404 Not Found" not in str(e):
-                    raise e
+                    raise
                 b = self._s3conn.create_bucket(self._bucket_name)
-            self._tlocal.s3bukt = b
+            self._tlocal.s3bukt = (b,time.time())
             return b
     _s3bukt = property(_s3bukt)
 
@@ -126,7 +156,8 @@ class S3FS(FS):
         self._tlocal = thread_local()
 
     def __str__(self):
-        return '<S3FS: %s:%s>' % (self._bucket_name,self._prefix)
+        args = (self.__class__.__name__,self._bucket_name,self._prefix)
+        return '<%s: %s:%s>' % args
 
     __repr__ = __str__
 
@@ -137,8 +168,20 @@ class S3FS(FS):
         s3path = self._prefix + path
         if s3path and s3path[-1] == self._separator:
             s3path = s3path[:-1]
+        if isinstance(s3path,unicode):
+            s3path = s3path.encode("utf8")
         return s3path
 
+    def _uns3path(self,s3path,roots3path=None):
+        """Get the local path for a file stored in S3.
+
+        This is essentially the opposite of self._s3path().
+        """
+        if roots3path is None:
+            roots3path = self._s3path("")
+        i = len(roots3path)
+        return s3path[i:]
+
     def _sync_key(self,k):
         """Synchronise on contents of the given key.
 
@@ -170,6 +213,10 @@ class S3FS(FS):
             key = self._s3bukt.new_key(key)
         if isinstance(contents,basestring):
             key.set_contents_from_string(contents)
+        elif hasattr(contents,"md5"):
+            hexmd5 = contents.md5
+            b64md5 = hexmd5.decode("hex").encode("base64").strip()
+            key.set_contents_from_file(contents,md5=(hexmd5,b64md5))
         else:
             try:
                 contents.seek(0)
@@ -180,13 +227,43 @@ class S3FS(FS):
                     tf.write(data)
                     data = contents.read(524288)
                 tf.seek(0)
-                contents = tf
-            key.set_contents_from_file(contents)
+                key.set_contents_from_file(tf)
+            else:
+                key.set_contents_from_file(contents)
         return self._sync_key(key)
 
-    def setcontents(self,path,contents):
+    def makepublic(self, path):
+        """Mark given path as publicly accessible using HTTP(S)"""
         s3path = self._s3path(path)
-        self._sync_set_contents(s3path,contents)
+        k = self._s3bukt.get_key(s3path)
+        k.make_public()
+        
+    def getpathurl(self, path, allow_none=False, expires=3600):
+        """Returns a url that corresponds to the given path."""
+        s3path = self._s3path(path)
+        k = self._s3bukt.get_key(s3path)
+
+        # Is there AllUsers group with READ permissions?
+        is_public = True in [grant.permission == 'READ' and \
+                grant.uri == 'http://acs.amazonaws.com/groups/global/AllUsers'
+                for grant in k.get_acl().acl.grants ]
+           
+        url = k.generate_url(expires, force_http=is_public)
+        
+        if url == None:
+            if not allow_none:
+                raise NoPathURLError(path=path)
+            return None
+
+        if is_public:
+            # Strip time token; it has no sense for public resource
+            url = url.split('?')[0]
+        
+        return url
+
+    def setcontents(self, path, data, chunk_size=64*1024):
+        s3path = self._s3path(path)
+        self._sync_set_contents(s3path, data)
 
     def open(self,path,mode="r"):
         """Open the named file in the given mode.
@@ -195,6 +272,8 @@ class S3FS(FS):
         so that it can be worked on efficiently.  Any changes made to the
         file are only sent back to S3 when the file is flushed or closed.
         """
+        if self.isdir(path):
+            raise ResourceInvalidError(path)
         s3path = self._s3path(path)
         # Truncate the file if requested
         if "w" in mode:
@@ -208,7 +287,14 @@ class S3FS(FS):
             if not self.isdir(dirname(path)):
                 raise ParentDirectoryMissingError(path)
             k = self._sync_set_contents(s3path,"")
-        return RemoteFileBuffer(self,path,mode,k)
+        #  Make sure nothing tries to read past end of socket data
+        f = LimitBytesFile(k.size,k,"r")
+        #  For streaming reads, return the key object directly
+        if mode == "r-":
+            return f
+        #  For everything else, use a RemoteFileBuffer.
+        #  This will take care of closing the socket when it's done.
+        return RemoteFileBuffer(self,path,mode,f)
 
     def exists(self,path):
         """Check whether a path exists."""
@@ -220,10 +306,10 @@ class S3FS(FS):
         ks = self._s3bukt.list(prefix=s3path,delimiter=self._separator)
         for k in ks:
             # A regular file
-            if k.name == s3path:
+            if _eq_utf8(k.name,s3path):
                 return True
             # A directory
-            if k.name == s3pathD:
+            if _eq_utf8(k.name,s3pathD):
                 return True
         return False
 
@@ -252,63 +338,100 @@ class S3FS(FS):
             return False
         k = self._s3bukt.get_key(s3path)
         if k is not None:
-          return True
+            return True
         return False
 
-    def listdir(self,path="./",wildcard=None,full=False,absolute=False,info=False,dirs_only=False,files_only=False):
+    def listdir(self,path="./",wildcard=None,full=False,absolute=False,
+                               dirs_only=False,files_only=False):
         """List contents of a directory."""
+        return list(self.ilistdir(path,wildcard,full,absolute,
+                                       dirs_only,files_only))
+
+    def listdirinfo(self,path="./",wildcard=None,full=False,absolute=False,
+                                   dirs_only=False,files_only=False):
+        return list(self.ilistdirinfo(path,wildcard,full,absolute,
+                                           dirs_only,files_only))
+
+    def ilistdir(self,path="./",wildcard=None,full=False,absolute=False,
+                                dirs_only=False,files_only=False):
+        """List contents of a directory."""
+        keys = self._iter_keys(path)
+        entries = self._filter_keys(path,keys,wildcard,full,absolute,
+                                         dirs_only,files_only)
+        return (nm for (nm,k) in entries)
+
+    def ilistdirinfo(self,path="./",wildcard=None,full=False,absolute=False,
+                                    dirs_only=False,files_only=False):
+        keys = self._iter_keys(path)
+        entries = self._filter_keys(path,keys,wildcard,full,absolute,
+                                         dirs_only,files_only)
+        return ((nm,self._get_key_info(k,nm)) for (nm,k) in entries)
+
+    def _iter_keys(self,path):
+        """Iterator over keys contained in the given directory.
+
+        This generator yields (name,key) pairs for each entry in the given
+        directory.  If the path is not a directory, it raises the approprate
+        error.
+        """
         s3path = self._s3path(path) + self._separator
         if s3path == "/":
             s3path = ""
-        i = len(s3path)
-        keys = []
         isDir = False
         for k in self._s3bukt.list(prefix=s3path,delimiter=self._separator):
             if not isDir:
                 isDir = True
             # Skip over the entry for the directory itself, if it exists
-            if k.name[i:] != "":
-                k.name = k.name[i:]
-                keys.append(k)
+            name = self._uns3path(k.name,s3path)
+            if name != "":
+                if not isinstance(name,unicode):
+                    name = name.decode("utf8")
+                if name.endswith(self._separator):
+                    name = name[:-1]
+                yield (name,k)
         if not isDir:
             if s3path != self._prefix:
                 if self.isfile(path):
-                    raise ResourceInvalidError(path,msg="that's not a directory: %(path)s")
+                    msg = "that's not a directory: %(path)s"
+                    raise ResourceInvalidError(path,msg=msg)
                 raise ResourceNotFoundError(path)
-        return self._listdir_helper(path,keys,wildcard,full,absolute,info,dirs_only,files_only)
 
-    def _listdir_helper(self,path,keys,wildcard,full,absolute,info,dirs_only,files_only):
-        """Modify listdir helper to avoid additional calls to the server."""
+    def _key_is_dir(self, k):
+        if isinstance(k,Prefix):
+            return True
+        if k.name.endswith(self._separator):
+            return True
+        return False
+
+    def _filter_keys(self,path,keys,wildcard,full,absolute,
+                               dirs_only,files_only):
+        """Filter out keys not matching the given criteria.
+
+        Given a (name,key) iterator as returned by _iter_keys, this method
+        applies the given filtering criteria and returns a filtered iterator.
+        """
+        sep = self._separator
         if dirs_only and files_only:
             raise ValueError("dirs_only and files_only can not both be True")
         if dirs_only:
-            keys = [k for k in keys if k.name.endswith(self._separator)]
+            keys = ((nm,k) for (nm,k) in keys if self._key_is_dir(k))
         elif files_only:
-            keys = [k for k in keys if not k.name.endswith(self._separator)]
-
-        for k in keys:
-            if k.name.endswith(self._separator):
-                k.name = k.name[:-1]
-            if type(path) is not unicode:
-                k.name = k.name.encode()
-
+            keys = ((nm,k) for (nm,k) in keys if not self._key_is_dir(k))
         if wildcard is not None:
-            keys = [k for k in keys if fnmatch.fnmatch(k.name, wildcard)]
-
+            if callable(wildcard):
+                keys = ((nm,k) for (nm,k) in keys if wildcard(nm))
+            else:
+                keys = ((nm,k) for (nm,k) in keys if fnmatch(nm,wildcard))
         if full:
-            entries = [relpath(pathjoin(path, k.name)) for k in keys]
+            return ((relpath(pathjoin(path, nm)),k) for (nm,k) in keys)
         elif absolute:
-            entries = [abspath(pathjoin(path, k.name)) for k in keys]
-        elif info:
-            entries = [self._get_key_info(k) for k in keys]
-        else:
-            entries = [k.name for k in keys]
-        return entries
+            return ((abspath(pathjoin(path, nm)),k) for (nm,k) in keys)
+        return keys
 
     def makedir(self,path,recursive=False,allow_recreate=False):
         """Create a directory at the given path.
 
-        The 'mode' argument is accepted for compatability with the standard
+        The 'mode' argument is accepted for compatibility with the standard
         FS interface, but is currently ignored.
         """
         s3path = self._s3path(path)
@@ -316,7 +439,9 @@ class S3FS(FS):
         if s3pathD == self._prefix:
             if allow_recreate:
                 return
-            raise DestinationExistsError(path, msg="Can not create a directory that already exists (try allow_recreate=True): %(path)s")
+            msg = "Can not create a directory that already exists"\
+                  " (try allow_recreate=True): %(path)s"
+            raise DestinationExistsError(path, msg=msg)
         s3pathP = self._s3path(dirname(path))
         if s3pathP:
             s3pathP = s3pathP + self._separator
@@ -329,22 +454,25 @@ class S3FS(FS):
         for k in ks:
             if not parentExists:
                 parentExists = True
-            if k.name == s3path:
+            if _eq_utf8(k.name,s3path):
                 # It's already a file
-                raise ResourceInvalidError(path, msg="Destination exists as a regular file: %(path)s")
-            if k.name == s3pathD:
+                msg = "Destination exists as a regular file: %(path)s"
+                raise ResourceInvalidError(path, msg=msg)
+            if _eq_utf8(k.name,s3pathD):
                 # It's already a directory
                 if allow_recreate:
                     return
-                raise DestinationExistsError(path, msg="Can not create a directory that already exists (try allow_recreate=True): %(path)s")
+                msg = "Can not create a directory that already exists"\
+                      " (try allow_recreate=True): %(path)s"
+                raise DestinationExistsError(path, msg=msg)
         # Create parent if required
         if not parentExists:
             if recursive:
                 self.makedir(dirname(path),recursive,allow_recreate)
             else:
-                raise ParentDirectoryMissingError(path, msg="Parent directory does not exist: %(path)s")
+                msg = "Parent directory does not exist: %(path)s"
+                raise ParentDirectoryMissingError(path, msg=msg)
         # Create an empty file representing the directory
-        # TODO: is there some standard scheme for representing empty dirs?
         self._sync_set_contents(s3pathD,"")
 
     def remove(self,path):
@@ -352,10 +480,11 @@ class S3FS(FS):
         s3path = self._s3path(path)
         ks = self._s3bukt.list(prefix=s3path,delimiter=self._separator)
         for k in ks:
-            if k.name == s3path:
+            if _eq_utf8(k.name,s3path):
                 break
-            if k.name.startswith(s3path + "/"):
-                raise ResourceInvalidError(path,msg="that's not a file: %(path)s")
+            if _startswith_utf8(k.name,s3path + "/"):
+                msg = "that's not a file: %(path)s"
+                raise ResourceInvalidError(path,msg=msg)
         else:
             raise ResourceNotFoundError(path)
         self._s3bukt.delete_key(s3path)
@@ -378,14 +507,16 @@ class S3FS(FS):
         found = False
         for k in ks:
             found = True
-            if k.name != s3path:
+            if not _eq_utf8(k.name,s3path):
                 if not force:
                     raise DirectoryNotEmptyError(path)
                 self._s3bukt.delete_key(k.name)
         if not found:
             if self.isfile(path):
-                raise ResourceInvalidError(path,msg="removedir() called on a regular file: %(path)s")
-            raise ResourceNotFoundError(path)
+                msg = "removedir() called on a regular file: %(path)s"
+                raise ResourceInvalidError(path,msg=msg)
+            if path not in ("","/"):
+                raise ResourceNotFoundError(path)
         self._s3bukt.delete_key(s3path)
         if recursive and path not in ("","/"):
             pdir = dirname(path)
@@ -397,7 +528,10 @@ class S3FS(FS):
     def rename(self,src,dst):
         """Rename the file at 'src' to 'dst'."""
         # Actually, in S3 'rename' is exactly the same as 'move'
-        self.move(src,dst)
+        if self.isfile(src):
+            self.move(src,dst)
+        else:
+            self.movedir(src,dst)
 
     def getinfo(self,path):
         s3path = self._s3path(path)
@@ -406,21 +540,31 @@ class S3FS(FS):
         else:
             k = self._s3bukt.get_key(s3path)
             if k is None:
-                k = self._s3bukt.get_key(s3path+"/")
-                if k is None:
+                ks = self._s3bukt.list(prefix=s3path,delimiter=self._separator)
+                for k in ks:
+                    if isinstance(k,Prefix):
+                        break
+                else:
                     raise ResourceNotFoundError(path)
-                k = Prefix(bucket=self._s3bukt,name=k.name)
-        return self._get_key_info(k)
+        return self._get_key_info(k,path)
 
-    def _get_key_info(self,key):
+    def _get_key_info(self,key,name=None):
         info = {}
-        info["name"] = basename(key.name)
-        if isinstance(key,Prefix):
+        if name is not None:
+            info["name"] = basename(name)
+        else:
+            info["name"] = basename(self._uns3key(k.name))
+        if self._key_is_dir(key):
             info["st_mode"] = 0700 | statinfo.S_IFDIR
         else:
             info["st_mode"] =  0700 | statinfo.S_IFREG
         if hasattr(key,"size"):
             info['size'] = int(key.size)
+        etag = getattr(key,"etag",None)
+        if etag is not None:
+            if isinstance(etag,unicode):
+               etag = etag.encode("utf8")
+            info['etag'] = etag.strip('"').strip("'")
         if hasattr(key,"last_modified"):
             # TODO: does S3 use any other formats?
             fmt = "%a, %d %b %Y %H:%M:%S %Z"
@@ -451,7 +595,7 @@ class S3FS(FS):
         dstOK = False
         for k in ks:
             # It exists as a regular file
-            if k.name == s3path_dst:
+            if _eq_utf8(k.name,s3path_dst):
                 if not overwrite:
                     raise DestinationExistsError(dst)
                 dstOK = True
@@ -459,23 +603,27 @@ class S3FS(FS):
             # Check if it refers to a directory.  If so, we copy *into* it.
             # Since S3 lists in lexicographic order, subsequent iterations
             # of the loop will check for the existence of the new filename.
-            if k.name == s3path_dstD:
+            if _eq_utf8(k.name,s3path_dstD):
                 nm = basename(src)
                 dst = pathjoin(dirname(dst),nm)
                 s3path_dst = s3path_dstD + nm
                 dstOK = True
         if not dstOK and not self.isdir(dirname(dst)):
-            raise ParentDirectoryMissingError(dst,msg="Destination directory does not exist: %(path)s")
+            msg = "Destination directory does not exist: %(path)s"
+            raise ParentDirectoryMissingError(dst,msg=msg)
         # OK, now we can copy the file.
         s3path_src = self._s3path(src)
         try:
             self._s3bukt.copy_key(s3path_dst,self._bucket_name,s3path_src)
         except S3ResponseError, e:
             if "404 Not Found" in str(e):
-                raise ResourceInvalidError(src, msg="Source is not a file: %(path)s")
+                msg = "Source is not a file: %(path)s"
+                raise ResourceInvalidError(src, msg=msg)
             raise e
         else:
             k = self._s3bukt.get_key(s3path_dst)
+            while k is None:
+                k = self._s3bukt.get_key(s3path_dst)
             self._sync_key(k)
 
     def move(self,src,dst,overwrite=False,chunk_size=16384):
@@ -483,6 +631,101 @@ class S3FS(FS):
         self.copy(src,dst,overwrite=overwrite)
         self._s3bukt.delete_key(self._s3path(src))
 
-    def get_total_size(self):
-        """Get total size of all files in this FS."""
-        return sum(k.size for k in self._s3bukt.list(prefix=self._prefix))
+    def walkfiles(self,
+              path="/",
+              wildcard=None,
+              dir_wildcard=None,
+              search="breadth",
+              ignore_errors=False ):
+        if search != "breadth" or dir_wildcard is not None:
+            args = (wildcard,dir_wildcard,search,ignore_errors)
+            for item in super(S3FS,self).walkfiles(path,*args):
+                yield item
+        else:
+            prefix = self._s3path(path)
+            for k in self._s3bukt.list(prefix=prefix): 
+                name = relpath(self._uns3path(k.name,prefix))
+                if name != "":
+                    if not isinstance(name,unicode):
+                        name = name.decode("utf8")
+                    if not k.name.endswith(self._separator):
+                        if wildcard is not None:
+                            if callable(wildcard):
+                                if not wildcard(basename(name)):
+                                    continue
+                            else:
+                                if not fnmatch(basename(name),wildcard):
+                                    continue
+                        yield pathjoin(path,name)
+
+
+    def walkinfo(self,
+              path="/",
+              wildcard=None,
+              dir_wildcard=None,
+              search="breadth",
+              ignore_errors=False ):
+        if search != "breadth" or dir_wildcard is not None:
+            args = (wildcard,dir_wildcard,search,ignore_errors)
+            for item in super(S3FS,self).walkfiles(path,*args):
+                yield (item,self.getinfo(item))
+        else:
+            prefix = self._s3path(path)
+            for k in self._s3bukt.list(prefix=prefix): 
+                name = relpath(self._uns3path(k.name,prefix))
+                if name != "":
+                    if not isinstance(name,unicode):
+                        name = name.decode("utf8")
+                    if wildcard is not None:
+                        if callable(wildcard):
+                            if not wildcard(basename(name)):
+                                continue
+                        else:
+                            if not fnmatch(basename(name),wildcard):
+                                continue
+                    yield (pathjoin(path,name),self._get_key_info(k,name))
+
+
+    def walkfilesinfo(self,
+              path="/",
+              wildcard=None,
+              dir_wildcard=None,
+              search="breadth",
+              ignore_errors=False ):
+        if search != "breadth" or dir_wildcard is not None:
+            args = (wildcard,dir_wildcard,search,ignore_errors)
+            for item in super(S3FS,self).walkfiles(path,*args):
+                yield (item,self.getinfo(item))
+        else:
+            prefix = self._s3path(path)
+            for k in self._s3bukt.list(prefix=prefix): 
+                name = relpath(self._uns3path(k.name,prefix))
+                if name != "":
+                    if not isinstance(name,unicode):
+                        name = name.decode("utf8")
+                    if not k.name.endswith(self._separator):
+                        if wildcard is not None:
+                            if callable(wildcard):
+                                if not wildcard(basename(name)):
+                                    continue
+                            else:
+                                if not fnmatch(basename(name),wildcard):
+                                    continue
+                        yield (pathjoin(path,name),self._get_key_info(k,name))
+
+
+
+def _eq_utf8(name1,name2):
+    if isinstance(name1,unicode):
+        name1 = name1.encode("utf8")
+    if isinstance(name2,unicode):
+        name2 = name2.encode("utf8")
+    return name1 == name2
+
+def _startswith_utf8(name1,name2):
+    if isinstance(name1,unicode):
+        name1 = name1.encode("utf8")
+    if isinstance(name2,unicode):
+        name2 = name2.encode("utf8")
+    return name1.startswith(name2)
+
