@@ -5,6 +5,8 @@
 
 """
 
+from __future__ import with_statement
+
 #  Send any output from the logging module to stdout, so it will
 #  be captured by nose and reported appropriately
 import sys
@@ -12,6 +14,9 @@ import logging
 logging.basicConfig(level=logging.ERROR, stream=sys.stdout)
 
 from fs.base import *
+from fs.path import *
+from fs.errors import *
+from fs.filelike import StringIO
 
 import datetime
 import unittest
@@ -34,6 +39,8 @@ class FSTestCases(object):
     To apply the tests to your own FS implementation, simply use FSTestCase
     as a mixin for your own unittest.TestCase subclass and have the setUp
     method set self.fs to an instance of your FS implementation.
+    
+    NB. The Filesystem being tested must have a capacity of at least 3MB.
 
     This class is designed as a mixin so that it's not detected by test
     loading tools such as nose.
@@ -43,6 +50,26 @@ class FSTestCases(object):
         """Check that a file exists within self.fs"""
         return self.fs.exists(p)
 
+    def test_meta(self):
+        """Checks getmeta / hasmeta are functioning"""
+        # getmeta / hasmeta are hard to test, since there is no way to validate
+        # the implementations response
+        meta_names = ["read_only",
+                      "network",
+                      "unicode_paths"]
+        stupid_meta = 'thismetashouldnotexist!"r$$%^&&*()_+'
+        self.assertRaises(NoMetaError, self.fs.getmeta, stupid_meta)
+        self.assertFalse(self.fs.hasmeta(stupid_meta))
+        self.assertEquals(None, self.fs.getmeta(stupid_meta, None))
+        self.assertEquals(3.14, self.fs.getmeta(stupid_meta, 3.14))
+        for meta_name in meta_names:
+            try:
+                meta = self.fs.getmeta(meta_name)
+                self.assertTrue(self.fs.hasmeta(meta_name))
+            except NoMetaError:
+                self.assertFalse(self.fs.hasmeta(meta_name))
+        
+
     def test_root_dir(self):
         self.assertTrue(self.fs.isdir(""))
         self.assertTrue(self.fs.isdir("/"))
@@ -66,6 +93,20 @@ class FSTestCases(object):
         repr(self.fs)
         self.assert_(hasattr(self.fs, 'desc'))
 
+    def test_open_on_directory(self):
+        self.fs.makedir("testdir")
+        try:
+            f = self.fs.open("testdir")
+        except ResourceInvalidError:
+            pass
+        except Exception:
+            raise
+            ecls = sys.exc_info()[0]
+            assert False, "%s raised instead of ResourceInvalidError" % (ecls,)
+        else:
+            f.close()
+            assert False, "ResourceInvalidError was not raised"
+      
     def test_writefile(self):
         self.assertRaises(ResourceNotFoundError,self.fs.open,"test1.txt")
         f = self.fs.open("test1.txt","w")
@@ -83,11 +124,39 @@ class FSTestCases(object):
         self.assertEquals(f.read(),"test file overwrite")
         f.close()
 
+    def test_setcontents(self):
+        #  setcontents() should accept both a string...
+        self.fs.setcontents("hello","world")
+        self.assertEquals(self.fs.getcontents("hello"),"world")
+        #  ...and a file-like object
+        self.fs.setcontents("hello",StringIO("to you, good sir!"))
+        self.assertEquals(self.fs.getcontents("hello"),"to you, good sir!")
+        #  setcontents() should accept both a string...
+        self.fs.setcontents("hello","world", chunk_size=2)
+        self.assertEquals(self.fs.getcontents("hello"),"world")
+        #  ...and a file-like object
+        self.fs.setcontents("hello",StringIO("to you, good sir!"), chunk_size=2)
+        self.assertEquals(self.fs.getcontents("hello"),"to you, good sir!")
+
+
+    def test_setcontents_async(self):
+        #  setcontents() should accept both a string...
+        self.fs.setcontents_async("hello", "world").wait()
+        self.assertEquals(self.fs.getcontents("hello"), "world")
+        #  ...and a file-like object
+        self.fs.setcontents_async("hello",StringIO("to you, good sir!")).wait()
+        self.assertEquals(self.fs.getcontents("hello"), "to you, good sir!")
+        self.fs.setcontents_async("hello", "world", chunk_size=2).wait()
+        self.assertEquals(self.fs.getcontents("hello"), "world")
+        #  ...and a file-like object
+        self.fs.setcontents_async("hello", StringIO("to you, good sir!"), chunk_size=2).wait()
+        self.assertEquals(self.fs.getcontents("hello"), "to you, good sir!")        
+
     def test_isdir_isfile(self):
         self.assertFalse(self.fs.exists("dir1"))
         self.assertFalse(self.fs.isdir("dir1"))
         self.assertFalse(self.fs.isfile("a.txt"))
-        self.fs.createfile("a.txt")
+        self.fs.setcontents("a.txt", '')
         self.assertFalse(self.fs.isdir("dir1"))
         self.assertTrue(self.fs.exists("a.txt"))
         self.assertTrue(self.fs.isfile("a.txt"))
@@ -103,10 +172,10 @@ class FSTestCases(object):
         def check_unicode(items):
             for item in items:
                 self.assertTrue(isinstance(item,unicode))
-        self.fs.createfile(u"a")
-        self.fs.createfile("b")
-        self.fs.createfile("foo")
-        self.fs.createfile("bar")
+        self.fs.setcontents(u"a", '')
+        self.fs.setcontents("b", '')
+        self.fs.setcontents("foo", '')
+        self.fs.setcontents("bar", '')
         # Test listing of the root directory
         d1 = self.fs.listdir()
         self.assertEqual(len(d1), 4)
@@ -127,10 +196,10 @@ class FSTestCases(object):
         # Create some deeper subdirectories, to make sure their
         # contents are not inadvertantly included
         self.fs.makedir("p/1/2/3",recursive=True)
-        self.fs.createfile("p/1/2/3/a")
-        self.fs.createfile("p/1/2/3/b")
-        self.fs.createfile("p/1/2/3/foo")
-        self.fs.createfile("p/1/2/3/bar")
+        self.fs.setcontents("p/1/2/3/a", '')
+        self.fs.setcontents("p/1/2/3/b", '')
+        self.fs.setcontents("p/1/2/3/foo", '')
+        self.fs.setcontents("p/1/2/3/bar", '')
         self.fs.makedir("q")
         # Test listing just files, just dirs, and wildcards
         dirs_only = self.fs.listdir(dirs_only=True)
@@ -160,12 +229,136 @@ class FSTestCases(object):
         self.assertRaises(ResourceNotFoundError,self.fs.listdir,"zebra")
         self.assertRaises(ResourceInvalidError,self.fs.listdir,"foo")
 
+    def test_listdirinfo(self):
+        def check_unicode(items):
+            for (nm,info) in items:
+                self.assertTrue(isinstance(nm,unicode))
+        def check_equal(items,target):
+            names = [nm for (nm,info) in items] 
+            self.assertEqual(sorted(names),sorted(target))
+        self.fs.setcontents(u"a", '')
+        self.fs.setcontents("b", '')
+        self.fs.setcontents("foo", '')
+        self.fs.setcontents("bar", '')
+        # Test listing of the root directory
+        d1 = self.fs.listdirinfo()
+        self.assertEqual(len(d1), 4)
+        check_equal(d1, [u"a", u"b", u"bar", u"foo"])
+        check_unicode(d1)
+        d1 = self.fs.listdirinfo("")
+        self.assertEqual(len(d1), 4)
+        check_equal(d1, [u"a", u"b", u"bar", u"foo"])
+        check_unicode(d1)
+        d1 = self.fs.listdirinfo("/")
+        self.assertEqual(len(d1), 4)
+        check_equal(d1, [u"a", u"b", u"bar", u"foo"])
+        check_unicode(d1)
+        # Test listing absolute paths
+        d2 = self.fs.listdirinfo(absolute=True)
+        self.assertEqual(len(d2), 4)
+        check_equal(d2, [u"/a", u"/b", u"/bar", u"/foo"])
+        check_unicode(d2)
+        # Create some deeper subdirectories, to make sure their
+        # contents are not inadvertantly included
+        self.fs.makedir("p/1/2/3",recursive=True)
+        self.fs.setcontents("p/1/2/3/a", '')
+        self.fs.setcontents("p/1/2/3/b", '')
+        self.fs.setcontents("p/1/2/3/foo", '')
+        self.fs.setcontents("p/1/2/3/bar", '')
+        self.fs.makedir("q")
+        # Test listing just files, just dirs, and wildcards
+        dirs_only = self.fs.listdirinfo(dirs_only=True)
+        files_only = self.fs.listdirinfo(files_only=True)
+        contains_a = self.fs.listdirinfo(wildcard="*a*")
+        check_equal(dirs_only, [u"p", u"q"])
+        check_equal(files_only, [u"a", u"b", u"bar", u"foo"])
+        check_equal(contains_a, [u"a",u"bar"])
+        check_unicode(dirs_only)
+        check_unicode(files_only)
+        check_unicode(contains_a)
+        # Test listing a subdirectory
+        d3 = self.fs.listdirinfo("p/1/2/3")
+        self.assertEqual(len(d3), 4)
+        check_equal(d3, [u"a", u"b", u"bar", u"foo"])
+        check_unicode(d3)
+        # Test listing a subdirectory with absoliute and full paths
+        d4 = self.fs.listdirinfo("p/1/2/3", absolute=True)
+        self.assertEqual(len(d4), 4)
+        check_equal(d4, [u"/p/1/2/3/a", u"/p/1/2/3/b", u"/p/1/2/3/bar", u"/p/1/2/3/foo"])
+        check_unicode(d4)
+        d4 = self.fs.listdirinfo("p/1/2/3", full=True)
+        self.assertEqual(len(d4), 4)
+        check_equal(d4, [u"p/1/2/3/a", u"p/1/2/3/b", u"p/1/2/3/bar", u"p/1/2/3/foo"])
+        check_unicode(d4)
+        # Test that appropriate errors are raised
+        self.assertRaises(ResourceNotFoundError,self.fs.listdirinfo,"zebra")
+        self.assertRaises(ResourceInvalidError,self.fs.listdirinfo,"foo")
+
+    def test_walk(self):
+        self.fs.setcontents('a.txt', 'hello')
+        self.fs.setcontents('b.txt', 'world')
+        self.fs.makeopendir('foo').setcontents('c', '123')
+        sorted_walk = sorted([(d,sorted(fs)) for (d,fs) in self.fs.walk()])
+        self.assertEquals(sorted_walk,
+                          [("/",["a.txt","b.txt"]),
+                           ("/foo",["c"])])
+        #  When searching breadth-first, shallow entries come first
+        found_a = False
+        for _,files in self.fs.walk(search="breadth"):
+            if "a.txt" in files:
+                found_a = True
+            if "c" in files:
+                break
+        assert found_a, "breadth search order was wrong"
+        #  When searching depth-first, deep entries come first
+        found_c = False
+        for _,files in self.fs.walk(search="depth"):
+            if "c" in files:
+                found_c = True
+            if "a.txt" in files:
+                break        
+        assert found_c, "depth search order was wrong: " + str(list(self.fs.walk(search="depth")))
+
+    def test_walk_wildcard(self):
+        self.fs.setcontents('a.txt', 'hello')
+        self.fs.setcontents('b.txt', 'world')
+        self.fs.makeopendir('foo').setcontents('c', '123')
+        self.fs.makeopendir('.svn').setcontents('ignored', '')
+        for dir_path, paths in self.fs.walk(wildcard='*.txt'):
+            for path in paths:
+                self.assert_(path.endswith('.txt'))
+        for dir_path, paths in self.fs.walk(wildcard=lambda fn:fn.endswith('.txt')):
+            for path in paths:
+                self.assert_(path.endswith('.txt'))
+
+    def test_walk_dir_wildcard(self):
+        self.fs.setcontents('a.txt', 'hello')
+        self.fs.setcontents('b.txt', 'world')
+        self.fs.makeopendir('foo').setcontents('c', '123')
+        self.fs.makeopendir('.svn').setcontents('ignored', '')
+        for dir_path, paths in self.fs.walk(dir_wildcard=lambda fn:not fn.endswith('.svn')):
+            for path in paths:
+                self.assert_('.svn' not in path)
+
+    def test_walkfiles(self):
+        self.fs.makeopendir('bar').setcontents('a.txt', '123')
+        self.fs.makeopendir('foo').setcontents('b', '123')
+        self.assertEquals(sorted(self.fs.walkfiles()),["/bar/a.txt","/foo/b"])
+        self.assertEquals(sorted(self.fs.walkfiles(dir_wildcard="*foo*")),["/foo/b"])
+        self.assertEquals(sorted(self.fs.walkfiles(wildcard="*.txt")),["/bar/a.txt"])
+
+    def test_walkdirs(self):
+        self.fs.makeopendir('bar').setcontents('a.txt', '123')
+        self.fs.makeopendir('foo').makeopendir("baz").setcontents('b', '123')
+        self.assertEquals(sorted(self.fs.walkdirs()),["/","/bar","/foo","/foo/baz"])
+        self.assertEquals(sorted(self.fs.walkdirs(wildcard="*foo*")),["/","/foo","/foo/baz"])
+
     def test_unicode(self):
         alpha = u"\N{GREEK SMALL LETTER ALPHA}"
         beta = u"\N{GREEK SMALL LETTER BETA}"
         self.fs.makedir(alpha)
-        self.fs.createfile(alpha+"/a")
-        self.fs.createfile(alpha+"/"+beta)
+        self.fs.setcontents(alpha+"/a", '')
+        self.fs.setcontents(alpha+"/"+beta, '')
         self.assertTrue(self.check(alpha))
         self.assertEquals(sorted(self.fs.listdir(alpha)),["a",beta])
 
@@ -182,18 +375,18 @@ class FSTestCases(object):
         self.assert_(check("a/b/child"))
         self.assertRaises(DestinationExistsError,self.fs.makedir,"/a/b")
         self.fs.makedir("/a/b",allow_recreate=True)
-        self.fs.createfile("/a/file")
+        self.fs.setcontents("/a/file", '')
         self.assertRaises(ResourceInvalidError,self.fs.makedir,"a/file")
 
     def test_remove(self):
-        self.fs.createfile("a.txt")
+        self.fs.setcontents("a.txt", '')
         self.assertTrue(self.check("a.txt"))
         self.fs.remove("a.txt")
         self.assertFalse(self.check("a.txt"))
         self.assertRaises(ResourceNotFoundError,self.fs.remove,"a.txt")
         self.fs.makedir("dir1")
         self.assertRaises(ResourceInvalidError,self.fs.remove,"dir1")
-        self.fs.createfile("/dir1/a.txt")
+        self.fs.setcontents("/dir1/a.txt", '')
         self.assertTrue(self.check("dir1/a.txt"))
         self.fs.remove("dir1/a.txt")
         self.assertFalse(self.check("/dir1/a.txt"))
@@ -219,9 +412,14 @@ class FSTestCases(object):
         self.assert_(not check("foo/bar/baz"))
         self.assert_(not check("foo/bar"))
         self.assert_(not check("foo"))
+        self.fs.makedir("foo/bar/baz", recursive=True)
+        self.fs.setcontents("foo/file.txt", "please don't delete me")
+        self.fs.removedir("foo/bar/baz", recursive=True)
+        self.assert_(not check("foo/bar/baz"))
+        self.assert_(not check("foo/bar"))
         #  Ensure that force=True works as expected
         self.fs.makedir("frollic/waggle", recursive=True)
-        self.fs.createfile("frollic/waddle.txt","waddlewaddlewaddle")
+        self.fs.setcontents("frollic/waddle.txt","waddlewaddlewaddle")
         self.assertRaises(DirectoryNotEmptyError,self.fs.removedir,"frollic")
         self.assertRaises(ResourceInvalidError,self.fs.removedir,"frollic/waddle.txt")
         self.fs.removedir("frollic",force=True)
@@ -241,14 +439,14 @@ class FSTestCases(object):
     def test_rename(self):
         check = self.check
         # test renaming a file in the same directory
-        self.fs.createfile("foo.txt","Hello, World!")
+        self.fs.setcontents("foo.txt","Hello, World!")
         self.assert_(check("foo.txt"))
         self.fs.rename("foo.txt", "bar.txt")
         self.assert_(check("bar.txt"))
         self.assert_(not check("foo.txt"))
         # test renaming a directory in the same directory
         self.fs.makedir("dir_a")
-        self.fs.createfile("dir_a/test.txt","testerific")
+        self.fs.setcontents("dir_a/test.txt","testerific")
         self.assert_(check("dir_a"))
         self.fs.rename("dir_a","dir_b")
         self.assert_(check("dir_b"))
@@ -265,7 +463,7 @@ class FSTestCases(object):
 
     def test_info(self):
         test_str = "Hello, World!"
-        self.fs.createfile("info.txt",test_str)
+        self.fs.setcontents("info.txt",test_str)
         info = self.fs.getinfo("info.txt")
         self.assertEqual(info['size'], len(test_str))
         self.fs.desc("info.txt")
@@ -274,7 +472,7 @@ class FSTestCases(object):
 
     def test_getsize(self):
         test_str = "*"*23
-        self.fs.createfile("info.txt",test_str)
+        self.fs.setcontents("info.txt",test_str)
         size = self.fs.getsize("info.txt")
         self.assertEqual(size, len(test_str))
 
@@ -282,7 +480,7 @@ class FSTestCases(object):
         check = self.check
         contents = "If the implementation is hard to explain, it's a bad idea."
         def makefile(path):
-            self.fs.createfile(path,contents)
+            self.fs.setcontents(path,contents)
         def checkcontents(path):
             check_contents = self.fs.getcontents(path)
             self.assertEqual(check_contents,contents)
@@ -315,8 +513,9 @@ class FSTestCases(object):
         check = self.check
         contents = "If the implementation is hard to explain, it's a bad idea."
         def makefile(path):
-            self.fs.createfile(path, contents)
+            self.fs.setcontents(path, contents)
 
+        self.assertRaises(ResourceNotFoundError,self.fs.movedir,"a","b")
         self.fs.makedir("a")
         self.fs.makedir("b")
         makefile("a/1.txt")
@@ -351,11 +550,15 @@ class FSTestCases(object):
         self.assert_(check("a/foo/bar/baz.txt"))
 
 
+    def test_cant_copy_from_os(self):
+        sys_executable = os.path.abspath(os.path.realpath(sys.executable))
+        self.assertRaises(FSError,self.fs.copy,sys_executable,"py.exe")
+
     def test_copyfile(self):
         check = self.check
         contents = "If the implementation is hard to explain, it's a bad idea."
         def makefile(path,contents=contents):
-            self.fs.createfile(path,contents)
+            self.fs.setcontents(path,contents)
         def checkcontents(path,contents=contents):
             check_contents = self.fs.getcontents(path)
             self.assertEqual(check_contents,contents)
@@ -365,9 +568,11 @@ class FSTestCases(object):
         makefile("foo/bar/a.txt")
         self.assert_(check("foo/bar/a.txt"))
         self.assert_(checkcontents("foo/bar/a.txt"))
+        #import rpdb2; rpdb2.start_embedded_debugger('password');
         self.fs.copy("foo/bar/a.txt", "foo/b.txt")
         self.assert_(check("foo/bar/a.txt"))
         self.assert_(check("foo/b.txt"))
+        self.assert_(checkcontents("foo/bar/a.txt"))
         self.assert_(checkcontents("foo/b.txt"))
 
         self.fs.copy("foo/b.txt", "c.txt")
@@ -388,7 +593,7 @@ class FSTestCases(object):
         check = self.check
         contents = "If the implementation is hard to explain, it's a bad idea."
         def makefile(path):
-            self.fs.createfile(path,contents)
+            self.fs.setcontents(path,contents)
         def checkcontents(path):
             check_contents = self.fs.getcontents(path)
             self.assertEqual(check_contents,contents)
@@ -427,7 +632,7 @@ class FSTestCases(object):
         check = self.check
         contents = "If the implementation is hard to explain, it's a bad idea."
         def makefile(path):
-            self.fs.createfile(path,contents)
+            self.fs.setcontents(path,contents)
 
         self.fs.makedir("a")
         makefile("a/1.txt")
@@ -503,6 +708,50 @@ class FSTestCases(object):
         f7.close()
         self.assertEqual(self.fs.getcontents("a.txt"), all_strings)
 
+    def test_truncate(self):
+        def checkcontents(path, check_contents):
+            read_contents = self.fs.getcontents(path)
+            self.assertEqual(read_contents,check_contents)
+            return read_contents == check_contents
+        self.fs.setcontents("hello","world")
+        checkcontents("hello","world")
+        self.fs.setcontents("hello","hi")
+        checkcontents("hello","hi")
+        self.fs.setcontents("hello","1234567890")
+        checkcontents("hello","1234567890")
+        with self.fs.open("hello","r+") as f:
+            f.truncate(7)
+        checkcontents("hello","1234567")
+        with self.fs.open("hello","r+") as f:
+            f.seek(5)
+            f.truncate()
+        checkcontents("hello","12345")
+
+    def test_truncate_to_larger_size(self):
+        
+        with self.fs.open("hello","w") as f:
+            f.truncate(30)
+        self.assertEquals(self.fs.getsize("hello"), 30)
+        
+        # Some file systems (FTPFS) don't support both reading and writing
+        if self.fs.getmeta('file.read_and_write', True):
+            with self.fs.open("hello","r+") as f:
+                f.seek(25)
+                f.write("123456")
+                
+            with self.fs.open("hello","r") as f:
+                f.seek(25)
+                self.assertEquals(f.read(),"123456")
+
+    def test_write_past_end_of_file(self):
+        if self.fs.getmeta('file.read_and_write', True):
+            with self.fs.open("write_at_end","w") as f:
+                f.seek(25)
+                f.write("EOF")
+            with self.fs.open("write_at_end","r") as f:
+                self.assertEquals(f.read(),"\x00"*25 + "EOF")
+
+
     def test_with_statement(self):
         #  This is a little tricky since 'with' is actually new syntax.
         #  We use eval() to make this method safe for old python versions.
@@ -527,20 +776,27 @@ class FSTestCases(object):
             self.assertEquals(self.fs.getcontents('f.txt'),contents)
 
     def test_pickling(self):
-        self.fs.createfile("test1","hello world")
-        fs2 = pickle.loads(pickle.dumps(self.fs))
-        self.assert_(fs2.isfile("test1"))
-        fs3 = pickle.loads(pickle.dumps(self.fs,-1))
-        self.assert_(fs3.isfile("test1"))
+        if self.fs.getmeta('pickle_contents', True):
+            self.fs.setcontents("test1","hello world")
+            fs2 = pickle.loads(pickle.dumps(self.fs))
+            self.assert_(fs2.isfile("test1"))
+            fs3 = pickle.loads(pickle.dumps(self.fs,-1))
+            self.assert_(fs3.isfile("test1"))
+        else:
+            # Just make sure it doesn't throw an exception
+            fs2 = pickle.loads(pickle.dumps(self.fs))
+            
 
     def test_big_file(self):
+        """Test handling of a big file (1MB)"""      
         chunk_size = 1024 * 256
         num_chunks = 4
         def chunk_stream():
             """Generate predictable-but-randomy binary content."""
             r = random.Random(0)
+            randint = r.randint
             for i in xrange(num_chunks):
-                c = "".join(chr(r.randint(0,255)) for j in xrange(chunk_size/8))
+                c = "".join(chr(randint(0,255)) for j in xrange(chunk_size/8))
                 yield c * 8
         f = self.fs.open("bigfile","wb")
         try:
@@ -569,7 +825,7 @@ class FSTestCases(object):
             return int(dts1) == int(dts2)       
         d1 = datetime.datetime(2010, 6, 20, 11, 0, 9, 987699)
         d2 = datetime.datetime(2010, 7, 5, 11, 0, 9, 500000)                        
-        self.fs.createfile('/dates.txt', 'check dates')        
+        self.fs.setcontents('/dates.txt', 'check dates')        
         # If the implementation supports settimes, check that the times
         # can be set and then retrieved
         try:
@@ -582,7 +838,7 @@ class FSTestCases(object):
             self.assertTrue(cmp_datetimes(d2, info['modified_time']))
 
 
-class ThreadingTestCases:
+class ThreadingTestCases(object):
     """Testcases for thread-safety of FS implementations."""
 
     #  These are either too slow to be worth repeating,
@@ -592,7 +848,9 @@ class ThreadingTestCases:
     __lock = threading.RLock()
 
     def _yield(self):
-        time.sleep(0.001)
+        #time.sleep(0.001)
+        # Yields without a delay
+        time.sleep(0)
 
     def _lock(self):
         self.__lock.acquire()
@@ -606,7 +864,9 @@ class ThreadingTestCases:
                 func()
             except Exception:
                 errors.append(sys.exc_info())
-        return threading.Thread(target=runThread)
+        thread = threading.Thread(target=runThread)
+        thread.daemon = True
+        return thread
 
     def _runThreads(self,*funcs):
         check_interval = sys.getcheckinterval()
@@ -623,7 +883,7 @@ class ThreadingTestCases:
         finally:
             sys.setcheckinterval(check_interval)
 
-    def test_setcontents(self):
+    def test_setcontents_threaded(self):
         def setcontents(name,contents):
             f = self.fs.open(name,"w")
             self._yield()
@@ -642,7 +902,7 @@ class ThreadingTestCases:
             self.assertEquals(self.fs.getcontents("thread2.txt"),c)
         self._runThreads(thread1,thread2)
 
-    def test_setcontents_samefile(self):
+    def test_setcontents_threaded_samefile(self):
         def setcontents(name,contents):
             f = self.fs.open(name,"w")
             self._yield()
@@ -673,9 +933,10 @@ class ThreadingTestCases:
             pass
 
     def test_cases_in_separate_dirs(self):
-        class TestCases_in_subdir(self.__class__):
+        class TestCases_in_subdir(self.__class__,unittest.TestCase):
             """Run all testcases against a subdir of self.fs"""
             def __init__(this,subdir):
+                super(TestCases_in_subdir, this).__init__("test_listdir")
                 this.subdir = subdir
                 for meth in dir(this):
                     if not meth.startswith("test_"):
