@@ -6,7 +6,9 @@ An FS wrapper class for blocking operations that would modify the FS.
 
 """
 
+from fs.base import NoDefaultMeta
 from fs.wrapfs import WrapFS
+from fs.errors import UnsupportedError, NoSysPathError
 
 class ReadOnlyFS(WrapFS):
     """ Makes a FS object read only. Any operation that could potentially modify
@@ -18,6 +20,16 @@ class ReadOnlyFS(WrapFS):
     accidental modification.
     
     """
+    
+    def getmeta(self, meta_name, default=NoDefaultMeta):
+        if meta_name == 'read_only':
+            return True
+        return self.wrapped_fs.getmeta(meta_name, default)
+        
+    def hasmeta(self, meta_name):
+        if meta_name == 'read_only':
+            return True
+        return self.wrapped_fs.hasmeta(meta_name)
         
     def getsyspath(self, path, allow_none=False):
         """ Doesn't technically modify the filesystem but could be used to work
@@ -44,4 +56,8 @@ class ReadOnlyFS(WrapFS):
     rename = _no_can_do
     setxattr = _no_can_do
     delattr = _no_can_do
-    
+    remove = _no_can_do
+    removedir = _no_can_do
+    settimes = _no_can_do
+    setcontents = _no_can_do
+    createfile = _no_can_do
